<copyright> TESLA mu.Invocation extensions.
    Written by <a href="mailto:tiggr@gerbil.org">Pieter J. Schoenmakers</a>

    Copyright &copy; 1998 Pieter J. Schoenmakers.

    This file is part of TESLA.  TESLA is distributed under the terms of
    the GNU General Public License, a copy of which can be found in the
    TESLA distribution; see the file COPYING.

    <id>$Id: Invocation.t,v 1.29 1998/07/24 15:01:48 tiggr Exp $</id>
    </copyright>

/******************** UnresolvedInvocation (tesla) ********************/

implementation class
mu.UnresolvedInvocation extension tesla: tesla

end;

implementation instance
mu.UnresolvedInvocation extension tesla
{
  <doc> Methods that are still a candidate for invocation.  </doc>
  MutableArray eligible_methods;

  <doc> The possible argument types.  Each element is a {Cons} of {Type}s
      while we are resolving.  </doc>
  MutableArray possible_arg_types;
}

<doc> Type the receiving invocation.  </doc>
(Expression, Cons) (resolved, possible_types)
  resolve (Cons, Type) (expected, convertible)
  context Meta context_type
  nested: (MutableIntArray, int) (indices, index)
{
  if (type != nil)
    {
      [[[ParseTom error self]
	 print ("internal: typed unresolved invocation")] nl];
      [[[ParseTom report self] print self] nl];

      return [super resolve (expected, convertible) context context_type
		    nested: (indices, index)];
    }

  // && [[expected car] isFullyDefinedType]
  Type convertible_at_index
    = !convertible ? nil : [convertible typeAt index in indices];
  int num_args = !arguments ? 0 : [arguments length];
  int i, num_em, read_i, write_i;
  Meta either;

  if (possible_arg_types != nil)
    {
      either = Meta ([[receiver type] actualSelf current_meta]);
      //      if (is_class)
      //	either = [either itsClass];
      num_em = !eligible_methods ? 0 : [eligible_methods length];

      if (expected != nil)
	{
	  /* Drop methods that no longer fit EXPECTED.  */
	  for (read_i = write_i = 0; read_i < num_em; read_i++)
	    {
	      CustomMethod m = eligible_methods[read_i];
	      Type t = [[m return_type] actualSelf either];

	      if (t != dynamic_type
		  && ![t elementOf expected covariant TRUE context context_type
			 nested: (indices, index)])
		continue;

	      if (write_i++ != read_i)
		eligible_methods[write_i - 1] = m;
	    }
	  if (read_i != write_i)
	    {
	      [eligible_methods truncate write_i];
	      num_em = write_i;
	    }
	}
    }
  else
    {
      /* First pass: we know nothing.  */
      possible_arg_types = [MutableObjectArray with num_args copies nil];

      receiver = [receiver resolve (top_type_list, nil) context context_type];
      either = Meta ([receiver type]);
      if (!either)
	return (self, nil);
      either = [either actualSelf current_meta];
      //      if (is_class)
      //	either = [either itsClass];

      eligible_methods = [either methods nil named name_parts with arguments
				 for self confined (is_super, confined)];
      if (!eligible_methods)
	{
	  [[[ParseTom error self]
	     print (either, " never responds to a method")] nl];
	  [[CustomMethod describe [ParseTom report self]
			 method (name_parts, arguments)] nl];
	  return (self, nil);
	}

      int j, ri, wi;
      num_em = [eligible_methods length];
      for (read_i = write_i = 0; read_i < num_em; read_i++)
	{
	  CustomMethod m = eligible_methods[read_i];
	  Type t = [[m return_type] actualSelf either];

	  /* If the return type can not match, skip this method.  */
	  if (t != dynamic_type && expected != nil
	      && ![t elementOf expected covariant TRUE context context_type
		     nested: (indices, index)])
	    continue;

	  /* If this is not an invocation to super, we can safely drop
             (duplicate) methods with truely identical types and name
             parts.  */
	  if (!is_super)
	    {
	      for (i = 0; i < write_i; i++)
		if ([m equal eligible_methods[i] names: YES types: YES
		       context either])
		  break;
	      if (i != write_i)
		continue;
	    }

	  /* Drop any methods from the ELIGIBLE_METHODS, up to WRITE_I,
             that are implemented by a superclass of the current method M.
             Drop M if it is implemented by a superclass of any of the
             those ELIGIBLE_METHODS up to WRITE_I.  */
	  Meta meta_m = [m meta];
	  for (ri = wi = 0; ri < write_i; ri++)
	    {
	      CustomMethod m2 = eligible_methods[ri];
	      if ([m equal m2 names: TRUE selectorTypes: TRUE
		     context either])
		{
		  Meta meta_m2 = [m2 meta];
		  if ([meta_m2 isProperSubclassOf meta_m])
		    {
		      /* M is overridden by M2.  Drop M.  */
		      break;
		    }
		  if ([meta_m isProperSubclassOf meta_m2])
		    {
		      /* M2 is overridden by M.  Drop M2.  */
		      continue;
		    }
		}
	      /* This one stays.  */
	      if (wi++ != ri)
		eligible_methods[wi - 1] = m2;
	    }
	  if (ri != write_i)
	    continue;
	  write_i = wi;

	  /* This one stays.  */
	  if (write_i++ != read_i)
	    eligible_methods[write_i - 1] = m;
	}

      [eligible_methods truncate write_i];
      num_em = write_i;

      if (num_em == 0)
	eligible_methods = nil;
      else if (num_args > 0)
	[self rebuildPossibleArgTypes (num_em, num_args) context either];
    }

  if (!eligible_methods)
    return (self, nil);

  boolean must_fully_resolve
    = (convertible != nil || (expected != nil && ![expected cdr]));
  boolean fully_typed, changes;

  for (;;)
    {
      long changed_arg_types = 0;
      if (num_args > 64)
	{
	  [[[ParseTom error self] print ("unimplemented: >64 arguments")] nl];
	  return (self, nil);
	}

      fully_typed = YES;

      /* Attempt to type the arguments.  */
      for (i = 0; i < num_args; i++)
	{
	  Cons possible2, possible = possible_arg_types[i];
	  Expression arg2, arg = arguments[i];

	  (arg2, possible2) = [arg resolve (possible, nil)
				   context context_type];

	  /* If the argument can not be typed, neither can we.  */
	  if (!possible2)
	    return (self, nil);

	  if (arg != arg2)
	    arguments[i] = arg = arg2;
	  if (![possible equal possible2])
	    {
	      possible_arg_types[i] = possible2;
	      changed_arg_types |= 1 << i;
	    }

	  if (fully_typed && ![arg type])
	    fully_typed = NO;
	}

      /* If nothing changed, we're done with typing and elimination.  */
      if (!changed_arg_types)
	break;

      /* Eliminate methods that do not fit the possible argument types of
         the arguments that changed.  */
      for (read_i = write_i = 0; read_i < num_em; read_i++)
	{
	  CustomMethod m = eligible_methods[read_i];

	  for (i = 0; i < num_args; i++)
	    if (changed_arg_types & (1 << i) != 0)
	      {
		Type t = [m formalArgumentType i in name_parts
			    context context_type];
		if (![t elementOf possible_arg_types[i] covariant FALSE
			context context_type])
		  break;
	      }

	  if (i != num_args)
	    continue;
	  /* This one stays.  */
	  if (write_i++ != read_i)
	    eligible_methods[write_i - 1] = m;
	}

      /* No methods eliminated: we're done typing and eliminating.  */
      if (read_i == write_i)
	break;
      [eligible_methods truncate write_i];
      num_em = write_i;

      if (!num_em)
	eligible_methods = nil;
      else
	{
	  /* Rebuild the changed possible argument types of all arguments,
	     now that some methods were eliminated.  */
	  for (i = 0; i < num_args; i++)
	    possible_arg_types[i] = nil;
	  if (num_em > 0 && num_args > 0)
	    [self rebuildPossibleArgTypes (num_em, num_args)
		  context context_type];
	}
    }

  if (!eligible_methods)
    return (self, nil);

  if ((must_fully_resolve || fully_typed) && num_em > 1)
    {
      /* Retain only those methods with the closest matching return type.  */
      if (must_fully_resolve && num_em > 1)
	{
	  int max_best = -1;
	  Type desired_type = !convertible ? [expected car] : convertible;
	  desired_type = [[desired_type typeAt index in indices]
			   actualSelf context_type];

	  for (read_i = write_i = 0; read_i < num_em; read_i++)
	    {
	      Method m = eligible_methods[read_i];
	      Type rt = [[m return_type] actualSelf either];
	      int d = [rt distanceToFormal desired_type];

	      /* D can be < 0 in a cast.  */
	      if (d >= 0 || max_best >= 0)
		{
		  if ((d >= 0 && d < max_best) || max_best == -1)
		    write_i = 0;
		  else if (d < 0 || d > max_best)
		    continue;

		  max_best = d;
		}

	      if (write_i++ != read_i)
		eligible_methods[write_i - 1] = m;
	    }
	  if (read_i != write_i)
	    {
	      [eligible_methods truncate write_i];
	      num_em = write_i;
	    }
	}

      /* Retain those methods with the highest number of currently fully
         matching arguments.  */
      if (num_em > 1 && num_args > 0)
	{
	  int d, max_best = -1;
	  for (read_i = write_i = 0; read_i < num_em; read_i++)
	    {
	      Method m = eligible_methods[read_i];

	      for (i = d = 0; i < num_args; i++)
		{
		  Type t = [arguments[i] type];
		  if (t != nil
		      && ![[t actualSelf context_type]
			    distanceToFormal [m formalArgumentType i
						in name_parts
						context context_type]])
		    d++;
		}
	      if (d > max_best)
		{
		  max_best = d;
		  write_i = 0;
		}
	      else if (d < max_best)
		continue;

	      if (write_i++ != read_i)
		eligible_methods[write_i - 1] = m;
	    }
	  if (read_i != write_i)
	    {
	      [eligible_methods truncate write_i];
	      num_em = write_i;
	    }
	}

      /* Prefer methods with less name parts.  */
      if (num_em > 1)
	{
	  int d, least = -1;

	  for (read_i = write_i = 0; read_i < num_em; read_i++)
	    {
	      Method m = eligible_methods[read_i];
	      int nnp = [m namePartsLength];

	      if (least == -1 || nnp < least)
		{
		  least = nnp;
		  write_i = 0;
		}
	      else if (nnp > least)
		continue;

	      if (write_i++ != read_i)
		eligible_methods[write_i - 1] = m;
	    }
	  if (read_i != write_i)
	    {
	      [eligible_methods truncate write_i];
	      num_em = write_i;
	    }
	}

      /* Drop those methods that involve dynamic typing, if any of them
         involve static typing.  */
      if (num_em > 1)
	{
	  boolean seen_dynamic;

	  for (read_i = write_i = 0; read_i < num_em; read_i++)
	    {
	      Method m = eligible_methods[read_i];
	      if ([m involvesDynamicType])
		{
		  boolean had_seen_dynamic = seen_dynamic;
		  seen_dynamic = TRUE;
		  if ((!had_seen_dynamic && write_i > 0)
		      || write_i != read_i)
		    continue;
		}
	      else if (write_i == read_i && seen_dynamic)
		write_i = 0;

	      if (write_i++ != read_i)
		eligible_methods[write_i - 1] = m;
	    }
	  if (read_i != write_i)
	    {
	      [eligible_methods truncate write_i];
	      num_em = write_i;
	    }
	}

      if (num_em > 1)
	{
	  /* We could be looking at a too loosely typed receiver, matching
             multiple methods.  If this is plain polymorphism and they all
             have the same selector, we can pick any of them.  I assume
             this does not happen for builtin methods.  */
	  Selector sel = [eligible_methods[0] actualSelector (nil, nil)];
	  for (read_i = 1; read_i < num_em; read_i++)
	    if (sel != [eligible_methods[read_i] actualSelector (nil, nil)])
	      break;

	  /* Set NUM_EM to 1 to have a method picked.  ELIGIBLE_METHODS
             will still contain the ones between which we could not
             choose.  */
	  if (read_i == num_em)
	    num_em = 1;
	}

//       if (num_em > 1)
// 	[self unimplemented cmd
// 	      message: "more choosing between multiple methods"];

      if (!num_em)
	eligible_methods = nil;
    }

  Cons possible;

  if (must_fully_resolve)
    if (num_em > 1)
      {
	// Ehhh...
	// Wed Apr 29 13:22:00 1998, tiggr@gerbil.org
	void;
      }
    else if (!fully_typed)
      {
	CustomMethod m = eligible_methods[0];
	for (i = 0; i < num_args; i++)
	  {
	    Expression arg = arguments[i];
	    if (![arg type])
	      {
		Expression arg2
		  = [arg resolve ([Cons with ([m formalArgumentType i
						 in name_parts
						 context context_type], nil)],
				  nil) context context_type];
		if (arg2 != arg)
		  arguments[i] = arg2;
	      }
	  }
	fully_typed = YES;
      }

  if (fully_typed && num_em == 1)
    {
      type = [[eligible_methods[0] return_type] actualSelf either];
      if (type != dynamic_type)
	possible = [Cons with (!expected ? type
			       : [type elementOf expected covariant YES
				       context context_type
				       nested: (indices, index)],
			       nil)];
      else if (!expected)
	{
	  type = [void_expr type];
	  possible = [Cons with (type, nil)];
	}
      else if (![expected cdr])
	{
	  type = [expected car];
	  possible = expected;
	}
      else
	[[[ParseTom error self] print "ambiguous dynamic return type"] nl];

      = ([self resolvedInvocation], possible);
    }
  else
    {
      /* Confine expected or create a list of possible types.  */
      for (i = 0; i < num_em; i++)
	{
	  Type t = [[eligible_methods[i] return_type] actualSelf either];
	  if (!expected)
	    possible = [t addToTypes possible];
	  else
	    possible = [t addToTypes possible ordered expected
			  nested: (indices, index)];
	}

      = (self, possible);
    }
}

/******************** private methods ********************/

<doc> Rebuild the possible argument types.  </doc>
protected void
  rebuildPossibleArgTypes (int, int) (num_methods, num_args)
		  context Meta context_type
{
  int im, ia;

  for (im = 0; im < num_methods; im++)
    {
      CustomMethod m = eligible_methods[im];
      for (ia = 0; ia < num_args; ia++)
	{
	  Type t = [m formalArgumentType ia in name_parts
		      context context_type];
	  possible_arg_types[ia] = [t addToTypes possible_arg_types[ia]];
	}
    }
}

<doc> Return a resolved invocation to replace the (now-resolved) receiving
    (previously) unresolved invocation.  The default implementation
    returns a {MethodInvocation}.  </doc>
mu.Invocation
  resolvedInvocation
{
  CustomMethod m = eligible_methods[0];
  boolean precise = [eligible_methods length] == 1;
  Meta superclass;

  if (is_super)
    {
      if (confined != nil)
	superclass = confined;
      else
	superclass = [[m extension] meta];

      Meta subclass = Meta ([[receiver type] actualSelf current_meta]);
      Meta direct = [subclass directSuperFor superclass];

      if (!direct)
	[[[ParseTom error self]
	   print ("ambiguous super reference to ", superclass)] nl];
      superclass = direct;
    }

  = [[MethodInvocation alloc self] initWithType type arguments arguments
      nameParts name_parts receiver receiver method m precise precise
      super superclass];
}

end;

/******************** UnresolvedOperatorInvocation ********************/

implementation class
UnresolvedOperatorInvocation extension tesla: mu.Invocation, tesla

end;

implementation instance
UnresolvedOperatorInvocation extension tesla

<doc> Return a resolved operator invocation to replace the (now-resolved)
    receiving (previously) unresolved invocation.  </doc>
OperatorInvocation
  resolvedInvocation
// pre && [eligible_methods[0] operator] == operator
{
  Method m = eligible_methods[0];
  = [[OperatorInvocation alloc self] initWithType type arguments arguments
				     method eligible_methods[0]
				     postfixed postfixed];
}

end;

/******************** MethodInvocation ********************/

implementation class
MethodInvocation: mu.Invocation, tesla

end;

implementation instance
MethodInvocation
{
  <doc> The template method we (possibly) are invoking.  Also good for
      retrieving the actual selector.  </doc>
  CustomMethod method;

  <doc> The receiver.  </doc>
  Expression receiver;

  <doc> The superclass.  </doc>
  public Meta superclass;
}

<doc> Designated initializer.  </doc>
id
  initWithType Type t
     arguments MutableArray a
     nameParts Indexed np
      receiver Expression rcv
	method CustomMethod m
       precise boolean precise
	 super Meta sm
{
  (receiver, superclass, method) = (rcv, sm, m);
  // Does this work for super?
  // Wed Jul  8 17:11:00 1998, tiggr@gerbil.org
  [method addDefaults a accordingTo np for self];
  = [super initWithType t arguments a];
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  [receiver coarseUsage f & ~EXPR_LHS | EXPR_RHS parent c body m];
  if (arguments != nil)
    {
      int i, n = [arguments length];
      for (i = 0; i < n; i++)
	[arguments[i] coarseUsage f & ~EXPR_LHS | EXPR_RHS parent c body m];
    }
}

Cexpr (result)
    precompile Cblock parent
	method Cblock body
{
  int i, n = !arguments ? 0 : [arguments length];
  Cexpr ercv = [receiver precompile parent method body];
  MutableArray eargs = !n ? nil : [MutableObjectArray withCapacity n];
  for (i = 0; i < n; i++)
    {
      Expression a = arguments[i];
      if (a != void_expr)
	[eargs add [a precompile parent method body]];
    }

  if (type != void_type)
    result = [type temporaryCexpr parent];
  Selector asel = [method actualSelector (arguments, type)];
  [asel noteUsage];

  Cexpr inv;
  if (!superclass)
    inv = [[Cdynmi alloc self] init (ercv, asel, eargs, result, self)];
  else
    {
      Meta subclass = Meta ([[receiver type] actualSelf current_meta]);
      SuperReference sr = [SuperReference from subclass to superclass];
      inv = [[Csupmi alloc self] init (sr, asel, eargs, result, self)];
    }

  [parent add inv];
  Selector fsel = [method formal_selector];
  if (asel != fsel)
    [fsel noteUseAsFormalFor asel];
}

<doc> Also emit the method and receiver.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s] print (" recv=", receiver, " method=", method)];
}

end;

/******************** OperatorInvocation ********************/

implementation class
OperatorInvocation: mu.Invocation, tesla

end;

implementation instance
OperatorInvocation
{
  <doc> The method we are invoking.  </doc>
  BuiltinMethod method;

  <doc> Iff {TRUE} we are postfixed.  This is only relevant for some unary
      operators.  </doc>
  boolean postfixed;
}

<doc> Designated initializer.  </doc>
id
  initWithType Type t
     arguments MutableArray a
	method BuiltinMethod m
     postfixed boolean pf
{
  (method, postfixed) = (m, pf);
  = [super initWithType t arguments a];
}

void
  coarseUsage int f
       parent Compound c
	 body MethodBody m
{
  [super coarseUsage f parent c body m];
  int i, n = [arguments length];
  for (i = 0; i < n; i++)
    [arguments[i] coarseUsage EXPR_RHS parent c body m];
}

Cexpr (result)
    precompile Cblock parent
	method Cblock body
{
  int op = [method operator];

  int n = [arguments length];
  if (expr_flags & EXPR_RHS != 0
      && !((op == OP_PP || op == OP_MM) && !postfixed))
    result = [type temporaryCexpr parent];

  Cexpr arg1 = [arguments[0] precompile parent method body];
  if (n == 2)
    if (op == OP_IMPLIES || op == OP_SC_OR || op == OP_SC_AND)
      {
	C tp, ep;
	Cblock alt = [[Cblock alloc] initWithParent parent];
	Cexpr altv = [arguments[1] precompile alt method body];
	[alt add [[Casgn alloc self] init (result, altv)]];

	if (op == OP_IMPLIES)
	  {
	    tp = alt;
	    ep = [[Casgn alloc self] init (result, [[Cconst alloc] init "1"])];
	  }
	else if (op == OP_SC_OR)
	  {
	    tp = [[Casgn alloc self] init (result, [[Cconst alloc] init "1"])];
	    ep = alt;
	  }
	else
	  {
	    tp = alt;
	    ep = [[Casgn alloc self] init (result, [[Cconst alloc] init "0"])];
	  }

	[parent add [[Cite alloc self] init (arg1, tp, ep)]];
      }
    else
      {
	String opname = operator_names[[method operator]];

	if (op == OP_LSR)
	  {
	    arg1 = [Clitcast with ("unsigned", arg1)];
	    opname = ">>";
	  }

	Cexpr arg2 = [arguments[1] precompile parent method body];

	if (result != nil)
	  [parent add [[Casgn alloc self] init (result,
			      [[Cbinop alloc] init (arg1, opname, arg2)])]];
      }
  else if (op == OP_PP || op == OP_MM)
    {
      Casgn mod = [[Casgn alloc self]
		    init (arg1, [[Cbinop alloc]
				  init (arg1, op == OP_PP ? "+" : "-",
					[[Cconst alloc] init "1"])])];
      if (!postfixed)
	[parent add mod];
      if (!result)
	result = arg1;
      else
	[parent add [[Casgn alloc self] init (result, arg1)]];
      if (postfixed)
	[parent add mod];
    }
  else if (result != nil)
    [parent add [[Casgn alloc self] init (result,
       [[Cmonop alloc] init (operator_names[[method operator]], arg1)])]];

}

<doc> Also emit the name of the operator.  </doc>
OutputStream
  writeFields OutputStream s
{
  = [[super writeFields s]
      print (" op=", String (operator_names[[method operator]]))];
}

end;
